// Copyright (C) 2017  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_GLOBAL_OPTIMIZATIOn_HEADER
#define DLIB_GLOBAL_OPTIMIZATIOn_HEADER

#include "global_optimization/upper_bound_function.h"
#include "global_optimization/global_function_search.h"
#include "global_optimization/find_max_global.h"

#endif // DLIB_GLOBAL_OPTIMIZATIOn_HEADER




